// Copyright (c) 2019, Dr. Bruno Guigas. All rights reserved. Use of this source
// code is governed by a BSD-style license that can be found in the LICENSE file.
import 'dart:typed_data';
import 'dart:math' as math;

/// This class provides functions to compute the mena value, the variance
/// and the standard deviation of the data in an array.
class Sigma {
  /// Returns the mean value of the numbers in [array]
  /// in the region [ixstart, ixend] (ixend exclusive)
  /// [ixstart] and [ixend] may be null, equivalent to the left/rightmost index.
  static double meanValue(Float64List array, int ixstart, int ixend) {
    if (ixstart == null) ixstart = 0;
    if (ixend == null) ixend = array.length - 1;
    double mean = 0.0;
    for (int i = ixstart; i < ixend; i++) {
      mean += array[i];
    }

    mean /= ixend - ixstart + 1;
    return mean;
  }

  /// Returns the variance of the numbers in [array] in the region [ixstart, ixend]
  /// (ixend exclusive): variance = mean(x**2) - (mean(x))**2
  /// [ixstart] and [ixend] may be null, equivalent to the left/rightmost index.
  static double variance(Float64List array, int ixstart, int ixend) {
    if (ixstart == null) ixstart = 0;
    if (ixend == null) ixend = array.length - 1;
    double mean = meanValue(array, ixstart, ixend);
    double variance = 0.0;
    double value;
    for (int i = ixstart; i < ixend; i++) {
      value = array[i];
      variance += (value - mean) * (value - mean);
    }

    variance /= ixend - ixstart + 1;

    return variance;
  }

  /// Returns the standard deviation of the numbers in [array] in the region
  /// [ixstart, ixend] (ixend exclusive): sigma = Sqrt(variance(arr))
  /// [ixstart] and [ixend] may be null, equivalent to the left/rightmost index.
  static double sigma(Float64List array, int ixstart, int ixend) {
    if (ixstart == null) ixstart = 0;
    if (ixend == null) ixend = array.length - 1;
    return math.sqrt(variance(array, ixstart, ixend));
  }
}
